require 'nn'

local ViterbiWrapper, parent = torch.class('nn.ViterbiWrapper','nn.Module')

local function cloneMLP(mlp)
   local mlpc
   if torch.typename(mlp) == 'nn.Sequential' then
      mlpc = nn.Sequential()
      for i=1,mlp:size() do
         mlpc:add( cloneMLP(mlp:get(i)) )
      end
   elseif torch.typename(mlp) == 'nn.Parallel' then
      mlpc = nn.Parallel()
      for i=1,#mlp.modules do
         mlpc:add( cloneMLP(mlp.modules[i]) )
      end
   else
      local f = torch.MemoryFile():binary()
      f:writeObject(mlp)
      f:seek(1)
      mlpc = f:readObject()
      f:close()

--       local f = torch.DiskFile('clone.bin', 'w'):binary()
--       f:writeObject(mlp)
--       f:close()
--       f = torch.DiskFile('clone.bin'):binary()
--       mlpc = f:readObject()
--       f:close()
   end
   if mlp.weight then
--      print('set weight in ' .. torch.typename(mlp))
      mlpc.weight:set(mlp.weight)
   end
   if mlp.bias then
--      print('set bias in ' .. torch.typename(mlp))
      mlpc.bias:set(mlp.bias)
   end
   return mlpc
end

function ViterbiWrapper:__init(mlp,nClass,maxSentenceLong)
   parent.__init(self)
   self.mlp = mlp
   self.nClass = nClass
   self.maxSentenceLong = maxSentenceLong or 0
   self.mlpfclones = {}
   for i=1,self.maxSentenceLong do
      self.mlpfclones[i] = cloneMLP(self.mlp)
      io.stdout:write('.')
      io.stdout:flush()
   end
   self.currentNerror = 0
end

function ViterbiWrapper:reset(stdv)
   self.mlp:reset(stdv)
end

function ViterbiWrapper:write(file)
   parent.write(self, file)
   file:writeObject(self.mlp)
   file:writeInt(self.nClass)
   file:writeInt(self.maxSentenceLong)
   file:writeInt(self.currentNerror)
end

function ViterbiWrapper:read(file)
   parent.read(self, file)
   self.mlp = file:readObject()
   self.nClass = file:readInt()
   self.maxSentenceLong = file:readInt()
   self.currentNerror = file:readInt()
   self.mlpfclones = {}
   -- for i=1,self.maxSentenceLong do
   --    self.mlpfclones[i] = cloneMLP(self.mlp)
   --    io.stdout:write('.')
   --    io.stdout:flush()
   -- end
   self.currentNerror = 0
end

function ViterbiWrapper:doClones()
   for i=1,self.maxSentenceLong do
      self.mlpfclones[i] = cloneMLP(self.mlp)
      io.stdout:write('.')
      io.stdout:flush()
   end
end

-- input is supposed to be a table of table, where each sub table is in the correct format
-- of the input of mlp
-- the outout is matrix where each column is the network output scores for corresponding input
function ViterbiWrapper:forward(input)
   local examples = input
   self.output:resize(self.nClass,#examples)
   self.output:fill(0)
   -- run the set through mlp
   for i,example in ipairs(examples) do
      local out = self.mlp:forward(example)
      self.output:select(2,i):copy(out)
   end
   return self.output
end

-- input is supposed to be a table of tables, where each sub table is in the correct format
-- of the input of mlp
-- gradOutput is supposed to be a matrix with each column representing the gradient from viterbi
-- per each error
function ViterbiWrapper:backward(input,gradOutput)
   local examples = input
   --self.gradInput:resizeAs(self.nClass,#examples)
   --print(gradOutput)
   local ce = 0
   for i=1, gradOutput:size(2) do
      if gradOutput:select(2,i):max() > 0 then
	 --print(i)
	 --table.foreach(input[i], print)
	 local ex = input[i]
	 ce = ce + 1
	 if self.maxSentenceLong == 0 then
	    self.mlp:forward(ex)
	    --self.mlp:zeroGradParameters()
--            print('ok, je backwarde sans le clone (!)')
	    self.mlp:backward(ex,gradOutput:select(2,i))
	 else
--            print('ok, je backwarde avec le clone')
	    self.mlpfclones[ce]:forward(ex)
	    self.mlpfclones[ce]:zeroGradParameters()
	    self.mlpfclones[ce]:backward(ex,gradOutput:select(2,i))
	 end
      end
   end
--   print('num of error in Viterbi = ' .. ce)
   self.currentNerror = ce
end

function ViterbiWrapper:zeroGradParameters()
   if self.maxSentenceLong == 0 then
      self.mlp:zeroGradParameters()
   else
      for i=1,self.currentNerror do
	 self.mlpfclones[i]:zeroGradParameters()
      end
   end
end

function ViterbiWrapper:updateParameters(learningRate)
   for i=1,self.currentNerror do
      if self.maxSentenceLong == 0 then
	 self.mlp:updateParameters(learningRate/self.currentNerror) -- note: we are in the loop (see the loop!) so currentNerror > 0!
      else
	 self.mlpfclones[i]:updateParameters(learningRate/self.currentNerror)
      end
   end
end
